#include "CyanLightControl.hpp"

#include "qthing.h"
#include "mqtt_common.h"

#include <string>
#include <cstdio>
#include <cstdlib>
#include <algorithm>

#include "esp_err.h"
#include "esp_log.h"
#include "driver/ledc.h"



const char *TAG = "CLC";
const std::string delimiter = ":";


float CyanLight::bytes2float(uint8_t *bytes) {
  float f;

  memcpy(&f, bytes, sizeof(f));

  return f;
}


CyanLight::Controller::Controller(uint8_t networkChannel, uint8_t channelsConfigured) : networkChannel{networkChannel} {

  this->channelsConfigured = std::max(channelsConfigured, CyanLight::MAX_CHANNELS);

  this->channels = new CyanLight::PWMChannel*[this->channelsConfigured];
  //this->channels = (CyanLight::PWMChannel **) malloc(this->channelsConfigured * sizeof(CyanLight::PWMChannel*));

  this->timer_cfg.duty_resolution = (ledc_timer_bit_t)this->resolution;
  this->timer_cfg.freq_hz = this->frequency;
  this->timer_cfg.speed_mode = LEDC_HIGH_SPEED_MODE;
  this->timer_cfg.timer_num = LEDC_TIMER_0;
  // Required for newer esp-idf:
  //ledc_timer.clk_cfg = LEDC_AUTO_CLK,             // Auto select the source clock
  ledc_timer_config(&this->timer_cfg);

  ledc_fade_func_install(0);


  for (uint8_t i = 0; i < channelsConfigured; i++) {
    this->channels[i] = new CyanLight::PWMChannel(i, CyanLight::channelGPIOS[i]);

    char topic[256];
    snprintf(topic, sizeof(topic), DEVICE_NAMESPACE "CyanLight/pwm/$%i", i);

    add_message_callback(topic, [&, i](std::string message) {

      this->setChannel(i, strtof(message.c_str(), NULL));

      // float f = strtof(message.c_str(), NULL);
      // ESP_LOGW(TAG, "i: %d   msg: %s   f: %f", i, message.c_str(), f);
    });
  }


  add_message_callback(DEVICE_NAMESPACE "CyanLight/frequency/set", [&](std::string message) {
    long int frq = strtol(message.c_str(), NULL, 0);
    this->setFrequency(frq);
  });

  add_message_callback(DEVICE_NAMESPACE "CyanLight/resolution/set", [&](std::string message) {
    long int res = strtol(message.c_str(), NULL, 0);
    this->setResolution(res);
  });

  // atomic setter for both values
  add_message_callback(DEVICE_NAMESPACE "CyanLight/frqres/set", [&](std::string message) {
    std::string::size_type found = message.find(delimiter);

    if (found == std::string::npos) {
      ESP_LOGW("CLC", "Invalid message format[ '%s' ]", message.c_str());
      return;
    }

    std::string _frq = message.substr(0, found);
    std::string _res = message.substr(found + delimiter.length());

    long int frq = strtol(_frq.c_str(), NULL, 0);
    long int res = strtol(_res.c_str(), NULL, 0);

    if (frq == 0 || res == 0) {
      ESP_LOGW("CLC", "Invalid message format[ '%s' ]", message.c_str());
      return;
    }

    this->setFrqRes(frq, res);
  });


  add_message_callback(DEVICE_NAMESPACE "CyanLight/frequency/get", [&](std::string ignored) {
    char tmp[16];
    snprintf(tmp, sizeof(tmp), "%i", this->getFrequency());
    publish_message(DEVICE_NAMESPACE "CyanLight/frequency", tmp);
  });

  add_message_callback(DEVICE_NAMESPACE "CyanLight/resolution/get", [&](std::string ignored) {
    char tmp[16];
    snprintf(tmp, sizeof(tmp), "%i", this->getResolution());
    publish_message(DEVICE_NAMESPACE "CyanLight/resolution", tmp);
  });

  // and for symmetric reasons: here comes the counterpart to the atomic frq-res-setter
  add_message_callback(DEVICE_NAMESPACE "CyanLight/frqres/get", [&](std::string ignored) {
    char tmp[32];
    snprintf(tmp, sizeof(tmp), "%i%s%i", this->getFrequency(), delimiter.c_str(), this->getResolution());
    publish_message(DEVICE_NAMESPACE "CyanLight/frqres", tmp);
  });
}


void CyanLight::Controller::setChannel(uint8_t channel, float value) {

  if (channel >= this->channelsConfigured) {
    ESP_LOGW(TAG, "ChannelID out of range: channel[ %i ]", channel);
    return;
  }

  if (value < 0.0f) value = 0.0f;
  if (value > 1.0f) value = 1.0f;

  uint32_t maxPWM = 1 << (this->resolution - 1);
  uint32_t pwm = (uint32_t)( value * (maxPWM - 1) );

  this->channels[channel]->setPWM(pwm);
}


bool CyanLight::Controller::setFrqRes(uint32_t frq_hz, uint8_t res_bits) {

  this->timer_cfg.duty_resolution = (ledc_timer_bit_t)res_bits;
  this->timer_cfg.freq_hz = frq_hz;

  if (ledc_timer_config(&this->timer_cfg) == ESP_OK) {
    this->frequency = frq_hz;
    this->resolution = res_bits;

    ESP_LOGI(TAG, "frequency[ %i Hz ]  resolution[ %i bit ]", frq_hz, res_bits);

    return true;

  } else {

    ESP_LOGW(TAG, "Invalid frequency[ %i Hz ] and resolution[ %i bit ] setting!", frq_hz, res_bits);
    return false;
  }
}

bool CyanLight::Controller::setFrequency(uint32_t frq_hz) {
  return this->setFrqRes(frq_hz, this->resolution);
}

uint32_t CyanLight::Controller::getFrequency() {
  return this->frequency;
}

bool CyanLight::Controller::setResolution(uint8_t res_bits) {
  return this->setFrqRes(this->frequency, res_bits);
}

uint8_t CyanLight::Controller::getResolution() {
  return this->resolution;
}
